#!/usr/local/bin/tclsh
#
# rwhois.tcl
#
# Koert Zeilstra - Network Solutions
# 1996.08.16
#

#
# Storage format of results = <query-result>
# <query-result> = { {<record> ...} {<error> ...} }
# <error> = <string>
# <record> = { <server> <port> {{<field-name> <value>} ... }
#

# rwhois_query
# Description:
#   Perform query and parse results, if desired
# Parameters:
#   server - server name (eg: "rwhois.internic.net")
#   port - port number (eg: 4321)
#   query - rwhois query string (eg: "domain internic.net")
#   config - list with config options: { {<option-name> <option-value>} ... }
#     <option-name> = parse | display | limit | cache | forward | language
#     <option-value> for <option-name>:
#       parse: yes | no
#       display: dump | swip
#       limit: <integer>
#       cache: on | off
#       forward: on | off
#       language: english
# Returns: <query-result>
#
proc rwhois_query {server port query config} {

    set parse yes
    set display dump
    set limit 20
    set cache off
    set forward off
    set language english
    set feedback ""

    foreach item $config {
	switch -- [lindex $item 0] {
	    parse { set parse [lindex $item 1] }
	    display { set display [lindex $item 1] }
	    limit { set limit [lindex $item 1] }
	    cache { set cache [lindex $item 1] }
	    forward { set forward [lindex $item 1] }
	    language { set language [lindex $item 1] }
	    feedback { set feedback [lindex $item 1] }
	}
    }

    if {$display != "dump"} { set parse no }
    if {$feedback != ""} { uplevel \#0 [list $feedback "Connecting to $server $port"] }

    set s [socket $server $port]
    puts $s "-RWhois V-1.0:001fff:0f:0f \[InterNIC B0.1\]"
    puts $s "-display $display"
    puts $s "-limit $limit"
    #puts $s "-cache $cache"
    #puts $s "-forward $forward"
    #puts $s "-language $language"
    puts $s "$query"
    flush $s
    set result [read -nonewline $s]
    close $s

    if {$parse != "yes"} {
	return $result
    } else {
	set lines [split $result "\n"]
	set record_list {}
	set record_lines {}
	set error_lines {}
	foreach line $lines {
	    #puts "line: $line"
	    if {$line == ""} {
		# end of record, put all lines in record
		#puts "---record done: $record_lines"
		if {$record_lines != {}} {
                    set parse_result [rwhois_parse_dump_record_lines $record_lines $server $port $query $config]
		    set result_type [lindex $parse_result 0]
                    if {$result_type == "record"} {
			lappend record_list [lindex $parse_result 1]
		    } else {
			set record_list [append $record_list [lindex [lindex $parse_result 1] 0]]
			set error_lines [append $error_lines [lindex [lindex $parse_result 1] 1]]
		    }
		}
		set record_lines {}
	    } else {
		if [string match "%*" $line] {
		    # line is special message, not query result
		    if [string match "%error*" $line] {
			lappend error_lines "$line\n"
		    }
		    if [string match "%info*" $line] {
			# Code needed to strip out %info message
		    }
		} else {
		    # line is part of query result
		    lappend record_lines $line
		}
	    }
	}

	# finish up last record end of record, put all lines in record

	if {$record_lines != {}} {
	    set parse_result [rwhois_parse_dump_record_lines $record_lines $server $port $query $config]
	    set result_type [lindex $parse_result 0]
	    if {$result_type == "record"} {
		lappend record_list [lindex $parse_result 1]
	    } else {
		set parse_result2 [lindex $parse_result 1]
		set new_errors [lindex $parse_result2 1]
		set record_list [append $record_list [lindex [lindex $parse_result 1] 0]]
		set error_lines $new_errors
	    }
	}

	#puts "---record_list: $record_list"
	#puts "---error_lines: $error_lines"

	return [list $record_list $error_lines]
    }
}

#
# rwhois_parse_dump_record_lines (internal use only)
# Description:
#   Parses the lines of the record dump and returns the result
# Parameters:
#   lines: list of strings
#   server: server name (eg: "rwhois.internic.net")
#   port: port number (eg: 4321)
#   query: query string (used for referrals)
#   config: config option list (used for referrals)
# Returns a result pair:
#   { "record" <record> } | { "referral" <query_result> }
#
proc rwhois_parse_dump_record_lines {lines server port query config} {
    set record_type ""
    set record_data {}
    set referral_result {}
    foreach line $lines {
        set i1 [string first ":" $line]
        set i2 -1
        if {$i1 != -1} {
	    set line2 [string range $line [expr $i1 + 1] end]
	    set i2 [string first ":" $line2]
	    if {$i2 != -1} {
		set schema [string range $line 0 [expr $i1 - 1]]
		set name [string range $line2 0 [expr $i2 - 1]]
		set value [string range $line2 [expr $i2 + 1] end]
                #puts "schema: $schema $name $value"
		if {$schema == "referral"} {
		    #puts "Referral: $name - $line"
		    if {$name == "Referral"} {
			#puts "Referral: $value"
			if [regexp "(.*):(.*):(.*)" $value x ref_server ref_port protocol] {
			    set referral_result [rwhois_query $ref_server $ref_port $query $config]
			}
		    }
		} else {
		    lappend record_data [list $name $value]
		}
	    }
	}
    }
    if {$referral_result != {}} {
	set result [list "referral" $referral_result]
    } else {
	set result [list "record" [list $server $port $record_data]]
    }
    #puts "***parse result: $result"
    return $result
}

#
# rwhois_parse_3tuple
# Description:
#   Parses tuple delimited with delimiter
# Parameters:
#   tuple: string with tuple (eg "domain:Handle:INTERNIC-DOM")
#   delimiter: string that delimits the fields
# Returns a list with the 3 fields:
#   { <field 1> <field 2> <field 3> }
proc rwhois_parse_3tuple {tuple delimiter} {
    set field1 ""
    set field2 ""
    set field3 ""
    set i1 [string first $delimiter $tuple]
    set i2 -1
    if {$i1 != -1} {
	set tuple2 [string range $tuple [expr $i1 + 1] end]
	set i2 [string first $delimiter $tuple2]
	if {$i2 != -1} {
	    set field1 [string range $tuple 0 [expr $i1 - 1]]
	    set field2 [string range $tuple2 0 [expr $i2 - 1]]
	    set field3 [string range $tuple2 [expr $i2 + 1] end]
	}
    }
    return [list $field1 $field2 $field3]
}

#
# rwhois_get_schema_names
# Description:
#   Retrieves the list of schemas
# Parameters:
#   server: server name (eg: "rwhois.internic.net")
#   port: port number (eg: 4321)
# Returns a list of schema names:
#   { <schema-name> ... }
proc rwhois_get_schema_names {server port} {
  set s [socket $server $port]
  puts $s "-RWhois V-1.0:001fff:0f:0f \[InterNIC B0.1\]"
  puts $s "-display dump"
  puts $s "-object"
  puts $s "-quit"
  flush $s
  set result [read -nonewline $s]
  close $s

  set lines [split $result "\n"]
  foreach line $lines {
    if [string match "%object *" $line] {
      set line2 [string trim [string range $line 7 end]]
      if {$line2 != ""} {
        set items [split $line2 ":"]
        set name [lindex $items 0]
        set attribute [lindex $items 1]
        set value [lindex $items 2]
        if {[info exists schema_array($name)] == 0} {
          set schema_array($name) {}
        }
        lappend schema_array($name) [list $attribute $value]
      }
    }
  }
  return [array names schema_array]
}

#
# rwhois_get_schema
# Description:
#   Get schema definition
# Parameters:
#   server - server name (eg: "rwhois.internic.net")
#   port - port number (eg: 4321)
#   schema_name - name of schema (eg: "domain")
# Returns a list of field definitions:
#   { {<field-name> <required> <multi-line>} ... }
#
proc rwhois_get_schema {server port schema_name} {
    set s [socket $server $port]
    puts $s "-RWhois V-1.0:001fff:0f:0f \[InterNIC B0.1\]"
    puts $s "-display dump"
    puts $s "-schema $schema_name"
    puts $s "-quit"
    flush $s
    set result [read -nonewline $s]
    close $s

    set schema_def {}
    set lines [split $result "\n"]
    set attribute_name ""
    set required ""
    set multi_line ""

    foreach line $lines {
	if [string match "%schema*" $line] {
	    set line2 [string trim [string range $line 8 end]]
	    if {$line2 == ""} {
		if {$attribute_name != ""} {
		    lappend schema_def [list $attribute_name $required $multi_line $type]
		}
		set attribute_name ""
		set type ""
		set required ""
		set multi_line ""
	    } else {
		set items [split $line2 ":"]
		set name [lindex $items 0]
		set attribute [lindex $items 1]
		set value [lindex $items 2]
		switch -- $attribute {
		    "attribute" { set attribute_name $value }
		    "type" { set type $value }
		    "required" { set required $value }
		    "mult-line" { set multi_line $value }
		}
	    }
	}
    }
    if {$attribute_name != ""} {
        lappend schema_def [list $attribute_name $required $multi_line $type]
    }
    return $schema_def
}

proc rwhois_get_status {server port} {
    set s [socket $server $port]
    puts $s "-RWhois V-1.0:001fff:0f:0f \[InterNIC B0.1\]"
    puts $s "-status"
    puts $s "-quit"
    flush $s
    set result [read -nonewline $s]
    close $s

    set status {}
    set lines [split $result "\n"]
    foreach line $lines {
	if [string match "%status *" $line] {
	    set i [string first ":" $line]
	    if {$i != -1} {
		set field [string trim [string range $line 8 [expr $i-1]]]
		set value [string trim [string range $line [expr $i+1] end]]
		lappend status [list $field $value]
	    }
	}
    }
    return $status
}

proc rwhois_encode_url {server port display restrictor query} {
    if {$port == 4321} {
	set hostport $server
    } else {
	set hostport "${server}:${port}"
    }

    set encoded_query ""
    foreach c [split $query {}] {
	switch -- $c {
	    " " { set c "%20" }
	}
	append encoded_query $c
    }

    set url "rwhois://$hostport/?${display}?${restrictor}?${encoded_query}"
    return $url
}

proc rwhois_decode_url {url} {
    set host ""
    set port 4321
    set display ""
    set restrictor ""
    set query ""
    if [regexp "rwhois://(.*)/\\?(.*)\\?(.*)\\?(.*)" $url hostport display restrictor encoded_query] {
puts "rwhois_decode_url: $hostport $display $restrictor $encoded_query"
	if {[string first ":" $hostport] == -1} {
	    set host hostport
	} else {
	    set items [split $hostport ":"]
	    set host [lindex $items 0]
	    set port [lindex $items 1]
	}
	set length [string length $encoded_query]
        for {set i 0} {$i < $length} {incr i} {
	}
    }
    return [list $host $port $display $restrictor $query]
}

proc rwhois_test {{query "internic.net"}} {
    set result [rwhois_query "dmeister" 4321 $query {}]
    set objects [lindex $result 0]
    set errors [lindex $result 1]
    foreach object $objects {
	puts "===object: $object"
    }
    foreach line $errors {
	puts "---error: $line"
    }
}

# rwhois_test
